@@ -1,5 +1,8 @@
+import sys
+import traceback
 from contextlib import contextmanager
 from collections import deque
+import logging
 from .function_call import FunctionCall
 from .setattr import Setattr
 from .exceptions import UnexpectedCall
@@ -7,6 +10,8 @@ from .exceptions import UnexpectedSetattr
 from .exceptions import ExpectedEventsNotFound
 from .utils import renumerate
 
+_logger = logging.getLogger("pyforge")
+
 class ForgeQueue(object):
     def __init__(self, forge):
         super(ForgeQueue, self).__init__()
@@ -18,7 +23,11 @@ class ForgeQueue(object):
             return 0
         return sum(len(group) for group in self._order_groups)
     def allow_whenever(self, queued_object):
+        self._discard(queued_object)
         self._whenever.add(queued_object)
+    def _discard(self, queued_object):
+        for group in self._order_groups:
+            group.discard(queued_object)
     def push_call(self, target, args, kwargs, caller_info):
         return self._push(FunctionCall(target, args, kwargs, caller_info))
     def push_setattr(self, target, name, value, caller_info):
@@ -36,8 +45,16 @@ class ForgeQueue(object):
             if popped is None:
                 popped = self._find_whenever_object(queued_object)
             if popped is None:
+                self._log_exception_context()
                 raise unexpected_class(self._get_replay_group().get_expected(), queued_object)
             return popped
+    def _log_exception_context(self):
+        exc_type, exc_value, exc_tb = sys.exc_info()
+        if exc_value is not None:
+            _logger.debug("In exception context:\n"
+                          "%s\n%s: %s",
+                          "".join(traceback.format_tb(exc_tb)),
+                          exc_type.__name__, exc_value)
     def _find_whenever_object(self, queued_object):
         for w in self._whenever:
             if w.matches(queued_object):
@@ -88,6 +105,8 @@ class OrderingGroup(object):
         return len(self._collection)
     def is_empty(self):
         return len(self) == 0
+    def discard(self, queued_object):
+        raise NotImplementedError() # pragma: no cover
 
 class OrderedGroup(OrderingGroup):
     def __init__(self):
@@ -102,6 +121,10 @@ class OrderedGroup(OrderingGroup):
         return None
     def get_expected(self):
         return self._collection[0] if self._collection else None
+    def discard(self, queued_object):
+        # only rebuild the queue if the queued object is in it...
+        if any(elem is queued_object for elem in self._collection):
+            self._collection = deque([elem for elem in self._collection if elem is not queued_object])
 
 class UnorderedGroup(OrderingGroup):
     def __init__(self):
@@ -115,3 +138,7 @@ class UnorderedGroup(OrderingGroup):
         return None
     def get_expected(self):
         return list(self._collection)
+    def discard(self, queued_object):
+        for index, elem in renumerate(self._collection):
+            if elem is queued_object:
+                self._collection.pop(index)
